#ifndef comms_h
#define comms_h

#ifndef general_h
#include "general.h"
#endif

#define BASE_PORT   5000
#define SERVER_PORT 5200

#define DATAMSG   98
#define SETUP_COMPLETE 99
#define ADMIND_ID -10	/* Must be negative and less than -1.      */

/* Types of text messages.                                         */
#define	SIGNON    1     /* Register with administration daemon.    */
#define	SIGNOFF	  2     /* Unregister with administration daemon.  */
#define ACCEPT    3     /* Signon accepted by daemon.              */
#define REJECT    4     /* Signon rejected by daemon.              */
#define NEGOTIATE 5	/* Establish connections to other clients 
			   in distributed system.                  */
#define	DATAREQ   6	/* Request for data from owner.            */
#define TOKEN     7     /* Required to access shared memory.       */
#define SCORE     8     /* Broadcast latest score information.     */
#define A_MISSILE 9     /* Broadcast missile information.          */
#define SYNC      10	/* Synchronous message.                    */
#define CLIENT_LK_ACQ 11	/* Client-side acquire lock.       */
#define MGR_LK_ACQ    12	/* Manager-side acquire lock.      */
#define MGR_LK_REL    13	/* " release lock.                 */


/* NOTE: accept_t and token_t make text-messages large, since a
   text-message is the union of the above message types. Both 
   accept_t and token_t are structures whose byte-sizes depend on
   the NUM_TEAMS constant. It would be more efficient to move
   token_t and accept_t into their own message-formats, instead
   of embedding them in text-messages. Either that or embed them
   in data messages which can be of variable length.               */

/* Match byte orders of differing machines.                        */
#define INSWAP(x)	((int)(ntohl((long)(x))))
#define OUTSWAP(x)	((int)(htonl((long)(x))))

#define HOST_LEN         64
#define NAME_LEN         64
#define MSG_LEN          128
#define MAX_IMAGE_SIZE   2048

/* Signoff flags.                                                  */
#define S_STOP 100
#define S_QUIT 101

/* Lock request types. */
#define LK_READ     900
#define LK_WRITE    901    

/* The first part of every message is a flag stating whether the 
   message is a text-only or data message.                         */
typedef enum {
  text,
  data,
  setup
} msg_t;


typedef enum {
  broadcast,
  multicast
} send_t;


typedef enum {
  locked,
  unlocked
} status_t;

typedef struct {
  int id;	
  char username [NAME_LEN];
  char hostname [HOST_LEN];
} user_t;


typedef struct {
  int goal_score;	       /* Current score for reaching goal. */
  int session_tot;             /* Current session score.           */
  int cumulative_tot;          /* Cumulative score for user's team 
				  over multiple sessions.          */
  user_t user;	               /* Data about user.                 */
} score_t;


typedef struct {
  int robot_number; /* Team robot responsible for firing missile. */
  int active;       /* Do we have an active missile that needs to 
		       be broadcast?                              */

  /* The following 'x' and 'y' members are coordinates associated 
     with the robot whose id is 'robot_number'. They are only 
     necessary with non-sequentially-consistent consistency 
     protocols. I feel ashamed at having to add them :-(          */
  int x;
  int y;
  int orientation;		/* Direction of fired missile.    */
} missile_t;


typedef struct {
  char text_info [MSG_LEN];
} text_t;


typedef struct writers *writers_p;

typedef struct writers {
  int client_id;
  unsigned long ts;
  writers_p next;
} writers_t;


typedef struct {
  int mgr;			/* Client acting as manager 
				   for lock.                      */
  int lock_id;			/* ID of lock.                    */
  int modified;			/* Set if locked data has been 
				   changed.                       */
  int type;			/* Type of lock requested.        */
  status_t status;		/* Client side: locked | unlocked.*/
  unsigned long ts;		/* Timestamp for locked object.   */
  int client_id;		/* Most recent client to update 
				   locked object.                 */
  writers_t *pending_writers;   /* List of pending clients       
				   wishing to acquire write lock
				   on object.                     */
} dlock_t;


/* Generic position structure for any object.*/
typedef struct {
  int x;            /* X position on battlefield.                 */
  int y;	    /* Y position on battlefield.                 */
  int orientation;  /* Orientation.                               */
  int object_type;  /* ENEMY_ROBOT, POS_OBS etc.                  */
} position_t;


typedef struct {
  int port;	    /* Callback port number.                      */
  char hostname[HOST_LEN];
  char username[NAME_LEN];
  int team_size;    /* Number of robots in team.                  */
} signon_t;


typedef struct {
  char reason[MSG_LEN]; /* Reason for signing off.                */
  int flag;		/* S_STOP, S_QUIT.                        */
} signoff_t;


/* Cut-down version of team_t maintained by each client.          */
typedef struct {
  int port;	       /* Callback port number.                   */
  int conn_id;         
  int registered;      /* TRUE if client has signed-on and is 
			  accepted; FALSE at sign-off.            */
  char username [NAME_LEN];
  char hostname [HOST_LEN];
  int team_size;
} client_t;


typedef struct {
  int id;	
  int goal_score;	      /* Initial score for reaching goal. */
  int num_clients;
  client_t client [NUM_TEAMS];
  int cp;			/* Consistency protocol to use.   */
} accept_t;


typedef struct {
  text_t text;	/* Textual reason for rejection.                  */
} reject_t;


typedef struct {
  int flag;   /* A valid request if non-zero. May be used to
	         specify requested object type.                   */
  int x ;     /* Top-left x coordinate of requested image object. */
  int y ;     /* Top-left y coordinate of requested image object. */
} datareq_t;


typedef struct {
  int conn_id;
} negotiate_t;


typedef struct {
  unsigned long ts;		/* Timestamp - May be vectorized 
				   in the future.                 */
  int sync_flag;		/* If set, have received SYNC msg.*/
  int resync_flag;		/* If set to FALSE, we shouldn't
				   resync with this client.       */
  int pending;
} sync_t;


typedef struct {
  int active[NUM_TEAMS]; /* Is this client still active when 
			    'token' is passed to another client?
			    When all clients see this array with 
			    all clients inactive, the token stops 
			    circulating the ring. */
  int conn_established[NUM_TEAMS]; 
  /* Have we opened symmetric connections to other clients yet?   */
} token_t;


typedef union {
  signon_t    signon;
  signoff_t   signoff;
  accept_t    accept;
  reject_t    reject;
  negotiate_t negotiate;
  datareq_t   datareq;
  token_t     token;
  dlock_t     dlock;
  score_t     score;
  missile_t   missile;
  sync_t      sync;
} info_t;


typedef struct {
  msg_t m;	       /* Indicates this is a text message.               */
  int id;              /* ID of sending machine.                          */
  int type;            /* Type of text message.                           */
  info_t info;
} textmsg_t;


typedef struct {
  int team_size;
  int robot_number;
  int robot_life;      /* Life points remaining for this robot.           */
  int killed_neg;      /* TRUE if this robot in team has walked over a 
			  negative obstacle.                              */
} rinfo_t;


typedef struct {
  msg_t m;	                /* Indicates this is a data message.     */
  int id;                       /* ID of sending machine.                */
  unsigned long ts;		/* Timestamp - may be vectored in future.*/
  position_t position;          /* Specifies top-left corner of 
				   image, and object type for 
				   consistency with world data 
				   structure.                            */
  rinfo_t rinfo;                /* Robot info if data is about a robot.  */
  int size;			/* # bytes of data to tag with message.  */
} datamsg_t;


typedef enum {
  empty,
  occupied
} stat_t;

typedef struct dbuf *databuf_p;
typedef struct dbuf {
  stat_t status;		/* Does this struct contain valid data?  */
  datamsg_t m;
  databuf_p next;		/* A bucket ptr for >= 2 buffered msgs.  */
  char *buf;			/* Ptr to variable-length data.          */
} databuf_t;


typedef struct {
  int life_points;     /* Life points remaining for this robot.           */
  int killed_neg;      /* TRUE if this robot in team has walked over a 
			  negative obstacle.                              */
  int orientation;     /* Direction robot is facing.                      */
  coordinate coord;    /* Location of a robot in a given team.            */ 
} robot_t;

/* As this code evolved team_t looks like client_t. Don't tell me, I 
   already know! */
typedef struct {
  int port;
  int team_number;
  int conn_id;         
  int registered;      /* TRUE if client has signed-on and is accepted; 
			  FALSE at sign-off.                              */
  int team_size;       /* Number of robots in team.                       */
  char username [NAME_LEN];
  char hostname [HOST_LEN];
  robot_t robot [MAX_ROBOTS];
} team_t;


void sendEvent (int (*func)(struct robot_args *), 
		struct robot_args *,
		int *exchange_list,
		int resync_flag,
		send_t how);
#endif




















