#include <stdio.h>

/* Define two generic schedulers for two classes of robots. Employ
   bounded buffers for both so that scheduler implementation can be left as
   flexible as possible. */

#include "general.h"
#include "scheduler.h"

task *buffer0 [MAX_BUFFER_SIZE]; /* EDF/SP bounded-buffer. */
task *buffer1 [MAX_BUFFER_SIZE]; /* FCFS_DD bounded-buffer. */

task *schedule_q0 [MAX_BUFFER_SIZE]; /* EDF/SP queue. */
task *schedule_q1 [MAX_BUFFER_SIZE]; /* FCFS_DD queue. (_DD = deadline dropping) */

static unsigned long next_start [NUM_SCHEDULERS];

void init_buffers_and_queues ()
{
  int i;
  
  for (i=0; i < MAX_BUFFER_SIZE; i++)
    {
      buffer0 [i] = NULL;
      buffer1 [i] = NULL;
      schedule_q0 [i] = NULL;
      schedule_q1 [i] = NULL;
    }
  
  for (i=0; i < NUM_SCHEDULERS; i++)
    {
      buffer [i].front = buffer[i].back = 0;
      buffer [i].status = EMPTY;
      queue [i].front = queue [i].back = queue [i].num_queued = 0;
      next_start [i] = current_time;
    }
}


/* Add new_arrival to back of appropriate bounded buffer. */
int add_to_buffer (task *new_arrival, int buffer_num, int front,
		   int *back, char *status)
{
  if (*back >= MAX_BUFFER_SIZE) {
    free (new_arrival->arg);
    free (new_arrival);
    return (EXCEEDED_BUFFER_LENGTH);
  }

  if ((*back == front) && (*status == OCCUPIED)) {
    free (new_arrival->arg);
    free (new_arrival);
    return (FULL_BUFFER);
  }
  
  switch (buffer_num) {
  case 0:
    buffer0 [*back] = new_arrival;
    break;
      
  case 1:
    buffer1 [*back] = new_arrival;
    break;
    
  default:
    return (NON_EXISTENT_BUFFER);
  }
  
  *back = (*back + 1) % MAX_BUFFER_SIZE;
  *status = OCCUPIED;
  return (SUCCESS);
}


/* Take item from front of appropriate buffer and form a new schedule. */
int add_to_schedule (int schedule_num, int *front, int back, char *status) {
  if (*front >= MAX_BUFFER_SIZE)
    return (EXCEEDED_BUFFER_LENGTH);
  
  if (queue[schedule_num].num_queued >= MAX_BUFFER_SIZE)
    return (EXCEEDED_BUFFER_LENGTH);
  
  if (*status == EMPTY)
      return (EMPTY_BUFFER);
  
  do {
    switch (schedule_num) {
    case 0:
      EDF_SP_sched (buffer0 [*front]);
      break;
	  
    case 1:
      FCFS_DD_sched (buffer1 [*front]);
      break;
      
    default:
      printf ("Invalid schedule number!\n");
      break;
    }
      
    *front = (*front + 1) % MAX_BUFFER_SIZE;
    if (*front == back)
      *status = EMPTY;
  }
  while ((*status != EMPTY) && 
	 (queue [schedule_num].num_queued < MAX_BUFFER_SIZE));
  
  if (*status == EMPTY)
    return (EMPTY_BUFFER);
  
  return (SUCCESS);
}


float value (task *current_task, unsigned long now) {
  float ramp_gradient;
  
  if (current_task->deadline == NO_DEADLINE)
    /* Task is not a dynamic priority task, 
       so does not have a value function assocated with it. */
    return (float) (current_task->initial_priority);
  
  if (current_task->deadline < (now + current_task->run_time))
    return 0.0; /* Deadline missed so task is now valueless! */
  
  switch (current_task->value_func) {
  case STEP_FUNCTION:
    return (float) (current_task->initial_priority);
    break;
      
  case RAMP_FUNCTION:
    ramp_gradient = current_task->initial_priority /
      (current_task->arrival_time - current_task->deadline);
    return ((ramp_gradient * (now - current_task->arrival_time)) + 
	    current_task->initial_priority);
    break;
  }
}


void EDF_SP_sched (task *new_task) {
  int i, j, k, min_pos, pred;
  float min_value_density;
  task *temp;
  static task *tentative_schedule [MAX_BUFFER_SIZE];
  unsigned long start_time_offset;
  int tentative_sched_count = 0;
  int finished=FALSE;
  int looped=FALSE; /* If front == back, when queue is full,
		       loop will never be executed without this baby! */
  
  /* Sort queue by EDF order to begin with.
     As working in simulated time, scheduling overhead is neglected.
     Hence, simple O(n) insertion sort TO AN ALREADY SORTED queue is
     used here...*/
  
  if (queue [0].num_queued == 0)
    {
      schedule_q0 [queue [0].back] = new_task;
      schedule_q0 [queue [0].back]->start_time = current_time;
    }
  
  else /* Loop until i == queue [0].front */
    {
      i = ((queue [0].back > 0) ? queue [0].back - 1 : MAX_BUFFER_SIZE - 1);
      
      for (; (i != queue [0].front) && !finished;
	     i = (i > 0 ? i - 1 : MAX_BUFFER_SIZE - 1))
	{
	  if (((schedule_q0 [i]->deadline > new_task->deadline) ||
	       (schedule_q0 [i]->deadline == NO_DEADLINE)) &&
	      (new_task->deadline != NO_DEADLINE))
	    {
	      temp = schedule_q0 [i];
	      schedule_q0 [i] = new_task;
	      schedule_q0 [(i+1) % MAX_BUFFER_SIZE] = temp;
	    }
	  
	  else
	    {
	      schedule_q0 [(i + 1) % MAX_BUFFER_SIZE] = new_task;
	      finished = TRUE;
	    }
	}
      
      /* And now the awkward boundary case when i == queue [0].front */
      if (!finished)
	{
	  if (((schedule_q0 [i]->deadline > new_task->deadline) ||
	       (schedule_q0 [i]->deadline == NO_DEADLINE)) &&
	      (new_task->deadline != NO_DEADLINE))
	    {
	      temp = schedule_q0 [i];
	      schedule_q0 [i] = new_task;
	      schedule_q0 [(i+1) % MAX_BUFFER_SIZE] = temp;
	    }
	  
	  else /* ?? This condition should have been taken care of by the above. */
	    {
	      schedule_q0 [(i + 1) % MAX_BUFFER_SIZE] = new_task;
	      finished = TRUE;
	    }
	}
    }
  
  
  queue [0].back = (queue [0].back + 1) % MAX_BUFFER_SIZE;
  queue [0].num_queued++;
  
  memset (tentative_schedule, 0, sizeof (task *)*MAX_BUFFER_SIZE);
  for (i = queue [0].front, j = queue [0].front;
       (i != queue [0].back) || !looped;
       i = (i + 1) % MAX_BUFFER_SIZE, j = (j + 1) % MAX_BUFFER_SIZE)
    {
      looped = TRUE;
      
      tentative_schedule [j] = schedule_q0 [i];
      tentative_sched_count++;
      if (j == queue [0].front)
	tentative_schedule [j]->start_time = current_time;
      
      else
	{
	  pred = (j > 0 ? j - 1 : MAX_BUFFER_SIZE - 1);
	  tentative_schedule [j]->start_time = 
	    tentative_schedule [pred]->start_time +
	    tentative_schedule [pred]->run_time;
	  
	  while (((tentative_schedule [j]->start_time +
		   tentative_schedule [j]->run_time) >
		  tentative_schedule [j]->deadline) &&
		 (tentative_schedule [j]->deadline != NO_DEADLINE) && 
		 (tentative_sched_count))
	    {
	      min_pos = queue [0].front;
	      min_value_density = value (tentative_schedule [min_pos], 
					 current_time) /
		tentative_schedule [min_pos]->run_time;
	      
	      for (k = min_pos; k != j; k = (k + 1) % MAX_BUFFER_SIZE)
		{
		  /* Find task with min value density. */
		  if ((value (tentative_schedule [k], current_time) /
		       tentative_schedule [k]->run_time) < min_value_density)
		    {
		      min_pos = k;
		      min_value_density = 
			value (tentative_schedule [k], current_time) /
			tentative_schedule [k]->run_time;
		    }
		}
	      
	      /* Now for the tedious case when k == j, posn. of the last element
		 added to our tentative schedule. */
	      
	      if ((value (tentative_schedule [k], current_time) /
		   tentative_schedule [k]->run_time) < min_value_density)
		{
		  min_pos = k;
		  min_value_density = value (tentative_schedule [k], 
					     current_time) /
		    tentative_schedule [k]->run_time;
		}

	      free (tentative_schedule [min_pos]->arg);
	      free (tentative_schedule [min_pos]);

	      start_time_offset = tentative_schedule [min_pos]->run_time;
	      
	      for (k = min_pos; k != j; k = (k + 1) % MAX_BUFFER_SIZE)
		{
		  /* Crunching array. */
		  tentative_schedule [(k + 1) % MAX_BUFFER_SIZE]->start_time -=
		    start_time_offset;
		  tentative_schedule [k] = 
		    tentative_schedule [(k + 1) % MAX_BUFFER_SIZE];
		}

	      queue [0].num_queued--;
	      if (--tentative_sched_count)
		j = ((j > 0) ? j - 1 : MAX_BUFFER_SIZE - 1); 
	    }
	}
    }
  queue [0].back = j;
  
  /* Now modify the true schedule as seen globally for queue 0. */
  for (i = 0; i < MAX_BUFFER_SIZE; i++)
    schedule_q0 [i] = tentative_schedule [i];
  
}


/* FCFS with dropping after deadline. Always add new tasks to back of queue
   and dispatch from the front. */

void FCFS_DD_sched (task *new_task)
{
  int i, j, pred;
  int looped=FALSE;
  unsigned long start_time_offset;
		
/* If front == back, when queue is full, 
   loop will never be executed without this baby! */
  
  schedule_q1 [queue [1].back] = new_task;
  pred = (queue [1].back > 0 ? queue [1].back - 1 : MAX_BUFFER_SIZE - 1);
  
  if (queue [1].num_queued == 0)
    schedule_q1 [queue [1].back]->start_time = current_time;
  else
    schedule_q1 [queue [1].back]->start_time = 
      schedule_q1 [pred]->start_time +
      schedule_q1 [pred]->run_time;
  
  queue [1].num_queued++;
  queue [1].back = (queue [1].back + 1) % MAX_BUFFER_SIZE;
  
  /* Remove any currently queued tasks that will now miss their deadlines. */
  for (i = queue [1].front; (i != queue [1].back) || !looped; 
       i = (i + 1) % MAX_BUFFER_SIZE)
    {
      looped = TRUE;
      
      if (((schedule_q1 [i]->deadline < 
	    (current_time + schedule_q1 [i]->run_time)) ||
	   (schedule_q1 [i]->deadline < (schedule_q1 [i]->start_time + 
					 schedule_q1 [i]->run_time)))
	  && (schedule_q1 [i]->deadline != NO_DEADLINE))
	{
	  start_time_offset = schedule_q1 [i]->run_time;

	  free (schedule_q1 [i]->arg);
	  free (schedule_q1 [i]);
	  
	  for (j = i; j != queue [1].back; j = (j + 1) % MAX_BUFFER_SIZE)
	    {
	      if ((j + 1) % MAX_BUFFER_SIZE != queue [1].back)
		{
		  schedule_q1 [(j+1) % MAX_BUFFER_SIZE]->start_time -= 
		    start_time_offset;  
		  schedule_q1 [j] = schedule_q1 [(j + 1) % MAX_BUFFER_SIZE];
		}
	    }
	  queue [1].num_queued--;
	  queue [1].back = (queue [1].back > 0 ? queue [1].back - 1 : 
			    MAX_BUFFER_SIZE - 1);
	  i = (i > 0 ? i - 1 : MAX_BUFFER_SIZE - 1);
	}
    }
  
}


/* A simple non-preemptive dispatcher. 
   Chooses task at front of appropriate scheduling queue
   and executes the related task function. */

int dispatch (int queue_num,
	      task **next_job)
{
  int return_val;
  
  if (next_start [queue_num] > current_time)
    return (CANNOT_DISPATCH); 
  /* Processor is still busy and system is non-preemptive! */
  
  if (queue [queue_num].num_queued) 
    /* Schedule queue is not empty: OK to dispatch. \n */
    {
      switch (queue_num)
	{
	case 0: /* EDF/SP queue. */
	  *next_job = schedule_q0 [queue [queue_num].front];
	  break;
	  
	case 1: /* FCFS_DD queue. */
	  *next_job = schedule_q1 [queue [queue_num].front];
	  break;
	  
	default: /* NON_EXISTENT_QUEUE */
	  printf ("Non-existent queue. \a\a\n");
	  break;
	}
      
      if ((*next_job)->start_time > current_time)
	{
	  printf ("Task start time greater than current time.\n");
	  return (CANNOT_DISPATCH); 
	  /* Start time of next task not yet reached. */
	}
      
      queue [queue_num].front = 
	(queue [queue_num].front + 1) % MAX_BUFFER_SIZE;
      
      next_start [queue_num] = current_time + (*next_job)->run_time;
      
      queue [queue_num].num_queued--;
      
      return_val = ((*next_job)->func) ((*next_job)->arg);
      
      return (return_val);
    }
  
  else
    return (CANNOT_DISPATCH); /* Schedule queue is empty. */
  
}






